"""Functions to download SRTM data from CGIAR website and generate a
TIF file for the desired coordinates using that data.
"""

import os
import math
import shutil
import zipfile
from typing import Optional, Union, Iterable, Tuple, Dict

import requests
import appdirs
import rasterio
import rasterio.merge
from rasterio.io import MemoryFile
from tqdm import tqdm


ZIP_BASE_URL = 'http://srtm.csi.cgiar.org/wp-content/uploads/files/srtm_5x5/TIFF/'
ZIP_FNAME = 'srtm_{x:02d}_{y:02d}.zip'
TIF_FNAME = 'srtm_{x:02d}_{y:02d}.tif'

TILE_X_BOUNDS = (1, 72)
TILE_Y_BOUNDS = (1, 24)
TILE_SHAPE = (6000, 6000)
SRTM_NODATA = -32768
#SRTM_NODATA = 65535
CACHE_DIR = appdirs.user_cache_dir('bother', appauthor=False)


def wrap_range(start: int, end: int, min_val: int = 1, max_val: int = 72) -> Iterable[int]:
    i = start
    if not ((min_val <= start <= max_val) and (min_val <= end <= max_val)):
        raise ValueError('start and end must each be between min_val and max_val.')
    while i != end:
        if i > max_val:
            i = min_val
        yield i
        i += 1

def is_cached(fname: str, cache_dir: str) -> bool:
    try:
        return fname in os.listdir(cache_dir)
    except FileNotFoundError:
        return False

def get_cached_files(cache_dir: str, ext: str = '') -> Iterable[str]:
    try:
        return list(filter(lambda f: f.endswith(ext), os.listdir(cache_dir)))
    except FileNotFoundError:
        return []

def get_extract_dir(cache_dir: str) -> str:
    return os.path.join(cache_dir, 'extracted')

def get_tif_fpath(x: int, y: int, cache_dir: str) -> str:
    return os.path.join(cache_dir, 'extracted', TIF_FNAME.format(x=x, y=y))

def download_zip(url: str, save_path: str, chunk_size: int = 1024):
    r = requests.get(url, stream=True)
    if not r.ok:
        if r.status_code == 404:
            # 404 can be returned if we try fetch a tile that corresponds to
            # a part of the world with no land, which is okay.
            return None
        else:
            r.raise_for_status()
    r.raise_for_status()
    total_size_in_bytes = int(r.headers.get('content-length', 0))
    
    # Write to a temporary file so that incompletely downloaded tiles are
    # not considered to have been cached
    temp_save_path = save_path + ".part"
    with tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True) as pbar:
        with open(temp_save_path, 'wb') as fd:
            for chunk in r.iter_content(chunk_size=chunk_size):
                pbar.update(len(chunk))
                fd.write(chunk)
    os.rename(temp_save_path, save_path)
    return save_path

def get_xy_components(lon: float, lat: float) -> Tuple[int, int]:
    """For a given longitude and latitude, returns the relevant
    components of the name of the zip file containing the SRTM data (for
    example "35" and "02" in "srtm_35_02.zip").
    """
    mod_lon = (int(math.floor(lon)) + 180) // 5 + 1
    mod_lat = (64 - int(math.floor(lat))) // 5
    return mod_lon, mod_lat

def get_all_xy_components(left: float, bottom: float, right: float, top: float) -> Iterable[Tuple[int, int]]:
    x_min, y_max = get_xy_components(left, bottom)
    x_max, y_min = get_xy_components(right, top)
    
    components = []
    
    for x in wrap_range(x_min, x_max+1, *TILE_X_BOUNDS):
        for y in wrap_range(y_min, y_max+1, *TILE_Y_BOUNDS):
            components.append((x, y))
    
    return components

#def get_raster_data(x: int, y: int, cache_dir) -> np.ndarray:
#    tif_fpath = get_tif_fpath(x, y, cache_dir)
#    if os.path.exists(tif_fpath):
#        with rasterio.open(tif_fpath) as src:
#            return src.read(1)
#    else:
#        memfile = rasterio.io.MemoryFile()
#        with memfile.open(driver='GTiff', count=1, width=6000, height=6000, dtype=np.uint16

def get_all_zip_fnames(left: float, bottom: float, right: float, top: float,
                     cache_dir: Optional[str] = None) -> Dict[Tuple[int, int], str]:
    """Get zip file URLs for all tiles relevant to the area within the
    given bounds.
    """
    fnames = {}
    for x, y in get_all_xy_components(left, bottom, right, top):
        fnames[(x, y)] = ZIP_FNAME.format(x=x, y=y)
    return fnames

def fetch_all_zips(fnames: Dict[Tuple[int, int], str], cache_dir: str) -> Dict[Tuple[int, int], str]:
    """Takes a dict mapping xy pairs to zip filenames, downloads the
    relevant files if they are not already in the cache, and returns
    a dict mapping each xy pair to the absolute paths of the relevant
    zip file.
    """
    
    fpaths = {}
    cached = get_cached_files(cache_dir, '.zip')
    for xy in fnames:
        fname = fnames[xy]
        fpath = os.path.join(cache_dir, fname)
        if fname not in cached:
            url = ZIP_BASE_URL+fname
            print(f'Downloading {fname} from {url}.')
            fpath = download_zip(url, fpath)
            if fpath is None:
                print(f'WARNING: Received HTTP 404 response when attempting to download {fname}.')
                print('This could mean that (1) this tile corresponds to a region of the earth '
                      'where there is no land; (2) this tile is outside the coverage area of the '
                      'SRTM data; or (3) there is some issue with the website (which may be '
                      'temporary).')
                print('We will assume the cause is (1) and will proceed accordingly.')
        else:
            print(f'{fname} was cached in {cache_dir}.')
        fpaths[xy] = fpath
    return fpaths

def unzip_all(zip_files: Iterable[str], cache_dir: str) -> str:
    extract_dir = get_extract_dir(cache_dir)
    os.makedirs(extract_dir, exist_ok=True)
    for fpath in zip_files:
        if fpath is not None:
            print(f'Extracting {fpath} to {extract_dir}.')
            with zipfile.ZipFile(fpath, 'r') as zf:
                zf.extractall(extract_dir)
    return extract_dir

def create_tif_file(left: float, bottom: float, right: float, top: float, to_file: Optional[str] = None,
                    cache_dir: str = CACHE_DIR, nodata: int = SRTM_NODATA) ->  Union[str, MemoryFile]:
    """Create a TIF file using SRTM data  for the box defined by left,
    bottom, right, top.  If to_file is provided, saves the resulting
    file to to_file and returns the path; otherwise, creates a
    rasterio.io.MemoryFile and returns that.
    """
    
    os.makedirs(cache_dir, exist_ok=True)
    
    xy = get_all_xy_components(left, bottom, right, top)
    zip_fnames = {}
    for x, y in xy:
        zip_fnames[(x, y)] = ZIP_FNAME.format(x=x, y=y)
    zip_fpaths = fetch_all_zips(zip_fnames, cache_dir)
    unzip_all(zip_fpaths.values(), cache_dir)
    srcs = []
    for x, y in xy:
        if zip_fpaths[(x, y)] is not None:
            srcs.append(rasterio.open(get_tif_fpath(x, y, cache_dir), 'r', nodata=nodata))
    print(f'Creating TIF file from following files: {[s.name for s in srcs]}.')
    #print(f'Heights are: {[s.height for s in srcs]}.')
    #print(f'Widths are: {[s.width for s in srcs]}.')
    profile = srcs[0].profile
    data, transform = rasterio.merge.merge(srcs, (left, bottom, right, top), nodata=nodata)
    for src in srcs:
        src.close()
    clear_cache(cache_dir, True)
    bands, height, width = data.shape   # No idea if this is the correct order for height and width, but they are both
                                        # the same so it doesn't matter in this case
    profile.update({
        'height': height,
        'width': width,
        'transform': transform
    })
    print(f'Created TIF file with dimensions {width}x{height}.') 
    if to_file:
        print(f'Writing TIF file to {to_file}.')
        with rasterio.open(to_file, 'w', **profile) as dst:
            dst.write(data)
        return to_file
    else:
        memfile = MemoryFile()
        with memfile.open(**profile) as dst:
            dst.write(data)
        return memfile

def clear_cache(cache_dir: str = CACHE_DIR, extracted_only: bool = False):
    if extracted_only:
        to_remove = get_extract_dir(cache_dir)
    else:
        to_remove = cache_dir
    try:
        shutil.rmtree(to_remove)
    except FileNotFoundError:
        pass



"""Functions to manipulate TIF files and convert them into heightmaps."""


import logging
import math
import time
import tempfile
import zipfile
from typing import Optional, Set, Tuple, List

import numpy as np
from PIL import Image

import rasterio
from rasterio.warp import calculate_default_transform, reproject, Resampling
from rasterio.io import MemoryFile

from bother_utils.srtm import SRTM_NODATA


WGS84 = 'EPSG:4326' # Mercator - The default CRS used in the STRM data

#def handle_nodata(memfile: MemoryFile, set_to: int = 0, nodata: int = SRTM_NODATA) -> MemoryFile:
#    
#    with memfile.open() as src:
#        
#        data = src.read(1)
#        data = np.where(data == nodata, set_to, data)
#        
#        dst_memfile = MemoryFile()
#        kwargs = src.profile.copy()
#        with dst_memfile.open(**kwargs) as dst:
#            dst.write(data, 1)
#
#        return dst_memfile
    
def remove_sea(memfile: MemoryFile, min_elev: int = 1) -> MemoryFile:
    """Offset elevation data so that lowest value is equal to min_elev.
    Useful for when real-world data includes land below sea level but no
    sea (eg, an in-land NL map).
    
    By default,min_elev is 1, which means you may need to use this in
    conjunction with raise_low_pixels to actually render as land.
    """
    
    with memfile.open() as src:
        
        data = src.read(1)
        offset = -(data.min() - min_elev)
        
        print(f'Increasing elevation by {offset}.')
        data += offset
        
        dst_memfile = MemoryFile()
        kwargs = src.profile.copy()
        with dst_memfile.open(**kwargs) as dst:
            dst.write(data, 1)

        return dst_memfile

def resample(memfile: MemoryFile, scale_factor: float) -> MemoryFile:
    """Resample raster by a factor of scale_factor.
        scale_factor > 1:  Upsample
        scale factor < 1:  Downsample
    """

    print(f'Resampling raster with scaling factor of {scale_factor}.')

    with memfile.open() as src:
        
        print(f'Source raster has shape {src.shape}.')
        
        # resample data to target shape
        height = int(src.height * scale_factor)
        width = int(src.width * scale_factor)

        data = src.read(
            out_shape=(src.count, height, width),
            resampling=Resampling.bilinear
        )
        # scale image transform
        transform = src.transform * src.transform.scale(width, height)

        kwargs = src.profile.copy()
        kwargs.update({
            'height': height,
            'width': width
        })
        dst_memfile = MemoryFile()
        with dst_memfile.open(**kwargs) as dst:
            for i in range(1, src.count+1):
                dst.write(data)
        
            print(f'Resampled raster has shape {dst.shape}.') 
        
        return dst_memfile

def reproject_raster(memfile: MemoryFile, dst_crs: str, src_crs: str = WGS84) -> MemoryFile:
    """Reproject raster with CRS src_crs to new CRS dst_crs."""
    
    print(f'Reprojecting raster from {src_crs} to {dst_crs}.')
    with memfile.open() as src:
        print(f'Source raster has shape {src.shape}.')
        transform, width, height = calculate_default_transform(src_crs, dst_crs, src.width, src.height, *src.bounds)
        kwargs = src.profile.copy()
        kwargs.update({
            'crs': dst_crs,
            'transform': transform,
            'width': width,
            'height': height
        })
        
        dst_memfile = MemoryFile()

        with dst_memfile.open(**kwargs) as dst:
            for i in range(1, src.count+1):
                reproject(
                    source=rasterio.band(src, i),
                    destination=rasterio.band(dst, i),
                    src_transform=src.transform,
                    src_crs=src_crs,
                    dst_transform=transform,
                    dst_crs=dst_crs,
                    resampling=Resampling.bilinear
                )
        
            print(f'Reprojected raster has shape {dst.shape}.')
        
        return dst_memfile

def get_lake(data: np.ndarray, row: int, col: int, checked: np.ndarray, min_size: int) -> Optional[Set[Tuple[int, int]]]:
    """Check if the pixel at data[row, col] belongs to a lake and, if so,
    return the lake as a set of points.
    
    We use a flood fill algorithm to detect contiguous sets of pixels
    that have exactly identical elevation.  This is similar to the
    algorithm (apparently) used by the MicroDEM software: see
    https://freegeographytools.com/2007/modifying-the-terrain-reflectance-display-in-microdem
    """
    elev = data[row, col]
    candidates = {(row,col)}
    lake = set()
    while candidates:
        (_row,_col) = candidates.pop()
        if checked[_row,_col]:
            continue
        try:
            candidate_elev = data[_row, _col]
        except IndexError:
            checked[_row, _col] = 1
            continue
        if candidate_elev == elev:
            lake.add((_row, _col))
            if row > 0:
                candidates.add((_row-1,_col))
            candidates.add((_row+1,_col))
            if col > 0:
                candidates.add((_row,_col-1))
            candidates.add((_row,_col+1))
        checked[_row, _col] = 1
    if len(lake) >= min_size:
        #print(f'Found lake of size {len(lake)}.')
        return lake

def get_all_lakes(data: np.ndarray, min_size: int, nodata: int = SRTM_NODATA) -> List[Set[Tuple[int, int]]]:
    """Find all lakes in the data.  A lake is defined as a contiguous
    region of at least min_size pixels of the exact same elevation."""

    height, width = data.shape
    checked = np.zeros((height+2, width+2))
    lakes = []
    for c in range(width):
        if (data[:,c] == nodata).all():
            checked[:,c] = 1
            continue
        for r in range(height):
            # Do basic checks before calling function, to avoid function call overhead
            if (not checked[r, c]) and (data[r, c] != nodata):
                lake = get_lake(data, r, c, checked, min_size)
                if lake is not None:
                    lakes.append(lake)
            checked[r, c] = 1
    return lakes

def set_lakes_to_elev(memfile: MemoryFile, min_lake_size: int, fill_lakes_as: int = None,
                      nodata: int = SRTM_NODATA) -> MemoryFile:
    """Find all lakes in the data for a raster and set the elevation of
    the relevant pixels to fill_lakes_as.
    """
    
    if fill_lakes_as is None:
        fill_lakes_as = nodata
    
    print(f'Finding lakes with minimum size of {min_lake_size} and setting elevation to {fill_lakes_as}.')
    with memfile.open() as src:
        data = src.read(1)
        lakes = get_all_lakes(data, min_lake_size)
        print(f'Found {len(lakes)} lakes.')
        for lake in lakes:
            for row, col in lake:
                data[row, col] = fill_lakes_as
        
        dst_memfile = MemoryFile()
        kwargs = src.profile.copy()
        with dst_memfile.open(**kwargs) as dst:
            dst.write(data, 1)

        return dst_memfile

def raise_undersea_land(memfile: MemoryFile, raise_to: int = 1, nodata: int = SRTM_NODATA):
    """Raise land with zero or negative elevation (ie, land that is at or
    below sea level) to raise_to. Probably shouldn't be called before
    set_lakes_to_elev, otherwise the newly raised land will be picked up
    as a lake by that function.
    """
    
    with memfile.open() as src:
        data = src.read(1)
        
        print(f'Raising pixels of elevation <= 0 to {raise_to} (ignoring NODATA).')
        data = np.where((data != nodata) & (data <= 0), raise_to, data)
    
        dst_memfile = MemoryFile()
        kwargs = src.profile.copy()
        with dst_memfile.open(**kwargs) as dst:
            dst.write(data, 1)
        
        return dst_memfile

def raise_low_pixels(memfile: MemoryFile, max_no_raise: float = 0.0, max_brightness: int = 255,
                     noisy: bool = False) -> MemoryFile:
    """Detect very low (but above sea level) elevations in a raster and
    increase the elevation of the relevant pixels such that, when the
    file is converted to a greyscale image, those pixels will be given
    a value of 1, rather than rounded down to 0.
    """
    
    with memfile.open() as src:
        data = src.read(1)
        # Floor all values at 0
        data *= (data > 0)
        max_elev = data.max()
        min_visible = math.ceil(max_elev / max_brightness) # Minimum value that will be rounded to 1 in a greyscale image
        print(f'Raising pixels of {max_no_raise} < elevation < {min_visible} to {min_visible}.')
        if noisy:
            # Add a small value to each affected pixel's elevation, which is proportionate to that pixel's original
            # elevation.  This ensures that fixed areas do not have a uniform elevation (unless they originally had
            # a uniform elevation), so that they are not picked up by Microdem as lakes.
            min_visible = (data / 10) + min_visible
        data = np.where((data > max_no_raise) & (data < min_visible), min_visible, data)
    
        dst_memfile = MemoryFile()
        kwargs = src.profile.copy()
        with dst_memfile.open(**kwargs) as dst:
            dst.write(data, 1)
        
        return dst_memfile

def to_png(memfile: MemoryFile, zero_floor: bool = False, max_brightness: int = 255, nodata: int = SRTM_NODATA):
    """Save raster as a greyscale PNG file to to_file.  If set_negative
    is set, any elevation values below zero are set to that value (which
    should be in the range 0-255).
    """
    
    print(f'Converting raster to PNG image.')
    with memfile.open() as src:
        data = src.read(1)
        data[data == nodata] = 0
        if zero_floor:
            data[data < 0] = 0
        max_elev = data.max()
        min_elev = data.min()
        scale_factor = max_brightness / (max_elev - min_elev)
        if min_elev > 0:
            # If everywhere on the map is above sea level, we should scale
            # such that the lowest parts of the map appear slightly above sea level.
            floor = min_elev + 1
        else:
            floor = min_elev
        data = ((data - floor) * scale_factor).astype(np.uint8)
        im = Image.fromarray(data, mode='L')
        width, height = im.size
        print(f'Image size is {width}x{height}.')
        return im

crop_modes = {'nw', 'n', 'ne', 'w', 'c', 'e', 'sw', 's', 'se'}
def crop_image(im: Image, width: int, height: int, mode: str) -> Image:
    """Crop an image to width x height. Where in the image to crop to is
    determined by mode.
    """
    
    mode = mode.lower()
    
    width = min(width, im.width)
    height = min(height, im.height)
    
    print(f'Cropping image to {width}x{height} using mode {mode}.')
    
    #       left top right bottom
    box =  [0, 0, 0, 0]
    
    if mode.startswith('n'):
        box[1] = 0
    elif mode.startswith('s'):
        box[1] = im.height - height
    else:
        box[1] = (im.height - height) // 2
    box[3] = box[1] + height
    
    if mode.endswith('w'):
        box[0] = 0
    elif mode.endswith('e'):
        box[0] = im.width - width
    else:
        box[0] = (im.width - width) // 2
    box[2] = box[0] + width
    

    # Image.crop chokes on large images unless we increase the max size
    size = width * height
    if size > Image.MAX_IMAGE_PIXELS:
        Image.MAX_IMAGE_PIXELS = size

    return im.crop(box)

def scale_image(im: Image, width: int, height: int) -> Image:
    """Scale an image to width x height."""
    print(f'Scaling image to {width}x{height}.')
    return im.resize((width, height))
        
def png_to_file(im: Image, to_file: str):
    """Save image to file."""
    
    print(f'Saving PNG image to {to_file}.')
    im.save(to_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
# sys.path.append(sys.argv[1])

import unittest
import os
import shutil

import numpy as np
from PIL import Image, ImageChops, ImageFile
ImageFile.LOAD_TRUNCATED_IMAGES = True

from bother_utils.heightmap import *
from bother_utils.srtm import *


EXAMPLES_DIR = os.path.join(os.path.dirname(__file__), 'examples')
TEST_DIR = os.path.join(os.path.dirname(__file__), 'test_data')

test_coords = {
    'mallorca': (39.195468, 2.272535, 39.998336, 3.584351),
    'ireland': (51.296094, -11.017116, 55.596810, -4.403352),
    'constance': (47.321556, 8.830936, 47.980224, 9.860904),
    'titicaca': (-16.970092, -70.246331, -15.032316, -68.334710),
    'alps': (46.408240, 9.555657, 46.975534, 10.378602),
    'germany_nw': (53.373069, 7.387423, 53.814826, 8.234738),
}

# We don't test this with the rest, because the raw image is too
# large to compare (PIL crashes)
vanc_oregon = (41.82, -129.946, 51.468, -116.125)

# No options provided
MALLORCA_TIF = os.path.join(EXAMPLES_DIR, 'mallorca.tif')
MALLORCA_PNG = os.path.join(EXAMPLES_DIR, 'mallorca.png')



class BotherTestCase(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        if not os.path.exists(TEST_DIR):
            os.mkdir(TEST_DIR)
        print(os.listdir())
    
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(TEST_DIR)
    
    def _assert_images_equal(self, im1, im2):
        np.testing.assert_array_equal(
            np.array(im1.getdata()),
            np.array(im2.getdata())
        )
    
    def _assert_image_files_equal(self, fpath1, fpath2):
        with Image.open(fpath1) as im1:
            with Image.open(fpath2) as im2:
                self._assert_images_equal(im1, im2)
    
    def test_01_srtm_tif(self):
        """Test the basic creation of TIF files."""
        
        for eg in test_coords:
            bottom, left, top, right = test_coords[eg]
            tif_file = os.path.join(TEST_DIR, f'{eg}.tif')
            eg_tif_file = os.path.join(EXAMPLES_DIR, f'{eg}.tif')
            create_tif_file(left, bottom, right, top, os.path.abspath(tif_file))
            self._assert_image_files_equal(tif_file, eg_tif_file)

    def test_02_resize(self):
        """Test basic cropping and scaling."""
        
        with open(os.path.join(TEST_DIR, 'mallorca.tif'), 'rb') as f:
            memfile = MemoryFile(f)
            with memfile.open() as src:
                data = src.read(1)
                shape = data.shape
            with memfile.open() as src:
                data = src.read(1)
            im = to_png(memfile, True, 255)
        im = crop_image(im, 1574, 787, 'c')
        im1 = scale_image(im, 2048, 1024)
        with Image.open(os.path.join(EXAMPLES_DIR, 'mallorca_no_processing_resized.png')) as im2:
            self._assert_images_equal(im1, im2)
    
    def test_03_downsample_lakes_raise_low_flat(self):
        """Test data resampling, lake detection, raising low pixels and flattening."""
        
        with open(os.path.join(EXAMPLES_DIR, 'ireland.tif'), 'rb') as f:
            memfile = MemoryFile(f)
            memfile = resample(memfile, 0.5)
            memfile = reproject_raster(memfile, dst_crs='EPSG:3857')
            memfile = set_lakes_to_elev(memfile, 100)
            memfile = raise_low_pixels(memfile)
            im1 = to_png(memfile)
        
        with Image.open(os.path.join(EXAMPLES_DIR, 'ireland_resampled_lakes_low_raised.png')) as im2:
            self._assert_images_equal(im1, im2)
        
        with open(os.path.join(EXAMPLES_DIR, 'ireland.tif'), 'rb') as f:
            memfile = MemoryFile(f)
            memfile = resample(memfile, 0.5)
            memfile = reproject_raster(memfile, dst_crs='EPSG:3857')
            memfile = set_lakes_to_elev(memfile, 100)
            memfile = raise_low_pixels(memfile, max_brightness=170)
            im1 = to_png(memfile, False, 170)
        with Image.open(os.path.join(EXAMPLES_DIR, 'ireland_resampled_lakes_low_raised_flat.png')) as im2:
            self._assert_images_equal(im1, im2)
            
    def test_04_lakes_all_above_sea_epsg_upsample(self):
        """Test a map where everything is above sea level (including one map where there are lakes).
        Also tests reprojection and upsampling of data.
        """
                
        with open(os.path.join(EXAMPLES_DIR, 'titicaca.tif'), 'rb') as f:
            memfile = MemoryFile(f)
            memfile = reproject_raster(memfile, dst_crs='EPSG:3395')
            memfile = set_lakes_to_elev(memfile, min_lake_size=80)
            im1 = to_png(memfile)
        with Image.open(os.path.join(EXAMPLES_DIR, 'titicaca_lakes_3395.png')) as im2:
            self._assert_images_equal(im1, im2)
            
        with open(os.path.join(EXAMPLES_DIR, 'titicaca.tif'), 'rb') as f:
            memfile = MemoryFile(f)
            memfile = reproject_raster(memfile, dst_crs='EPSG:4326')
            memfile = set_lakes_to_elev(memfile, min_lake_size=80)
            im1 = to_png(memfile)
        with Image.open(os.path.join(EXAMPLES_DIR, 'titicaca_lakes_4326.png')) as im2:
            self._assert_images_equal(im1, im2)
        
        with open(os.path.join(EXAMPLES_DIR, 'titicaca.tif'), 'rb') as f:
            memfile = MemoryFile(f)
            memfile = set_lakes_to_elev(memfile, min_lake_size=80)
            im1 = to_png(memfile)
        with Image.open(os.path.join(EXAMPLES_DIR, 'titicaca_lakes_4326.png')) as im2:
            self._assert_images_equal(im1, im2)
        
        with open(os.path.join(EXAMPLES_DIR, 'alps.tif'), 'rb') as f:
            memfile = MemoryFile(f)
            memfile = reproject_raster(memfile, dst_crs='EPSG:3857')
            im1 = to_png(memfile)
        with Image.open(os.path.join(EXAMPLES_DIR, 'alps.png')) as im2:
            self._assert_images_equal(im1, im2)
    
    def test_05_nested_cache_dir(self):
        """Test cache directory creation when multiple levels of the cache path do not exist."""
        
        bottom, left, top, right = test_coords["mallorca"]
        cache_dir = os.path.join(TEST_DIR, "level1", "level2")
        tif_file = os.path.join(TEST_DIR, f'mallorca.tif')
        eg_tif_file = os.path.join(EXAMPLES_DIR, f'mallorca.tif')
        create_tif_file(left, bottom, right, top, os.path.abspath(tif_file), cache_dir=cache_dir)
        self._assert_image_files_equal(tif_file, eg_tif_file)
    
    def test_06_raise_undersea(self):
        """Test raising undersea pixels."""
        
        bottom, left, top, right = test_coords['germany_nw']
        with open(os.path.join(EXAMPLES_DIR, 'germany_nw.tif'), 'rb') as f:
            memfile = MemoryFile(f)
            memfile = reproject_raster(memfile, dst_crs='EPSG:3857')
            memfile = raise_undersea_land(memfile, 1)
            memfile = raise_low_pixels(memfile, 0)
            im1 = to_png(memfile)
        with Image.open(os.path.join(EXAMPLES_DIR, 'germany_nw_undersea_raised.png')) as im2:
            self._assert_images_equal(im1, im2)

    def test_07_water_tile(self):
        """Test coverage of tile areas with no land."""
        
        bottom, left, top, right = vanc_oregon
        tif_file = os.path.join(TEST_DIR, 'vanc_oregon.tif')
        create_tif_file(left, bottom, right, top, os.path.abspath(tif_file))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

scripts = ['bother']

if os.name == 'nt':
    scripts.append(os.path.join('scripts', 'bother.bat'))
    
setup(
    name='Bother',
    version='1.1',
    packages=find_packages(),
    scripts=scripts,
    install_requires=['numpy', 'rasterio', 'pillow', 'requests', 'appdirs', 'tqdm'],
    package_data={},
    author='bunburya',
    author_email='',
    description='A script to produce heightmaps (primarily for OpenTTD) using real-world elevation data.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='srtm elevation heightmap openttd',
    url='https://github.com/bunburya/bother',
    classifiers=[
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: MIT License'
    ]
)
